a {
  text-decoration: none;
  color: #069;
}

a:hover {
  text-decoration: underline;
}

a:visited {
  color: #036;
}

body {
  font-family: Verdana, Helvetica, arial, freesans, clean, sans-serif;
  font-size: 10pt;
  width: 960px;
  margin-left: auto;
  margin-right: auto;
  background-color: #FFF;
  color: #444;
}

code {
  font-size: 10pt;
}

.table-contents table td,
.table-contents table th,
.navheader table td,
.navfooter table td {
  font-size: 10pt;
}

.navfooter hr, 
.navheader hr {
  border: 0px;
  border-bottom: 1px solid #fc390e; /* jetty orange */
}

.navfooter a,
.navheader a {
  font-weight: bold;
  color: #fc390e;
}

h1, h2, h3, h4, h5, h6 {
  font-weight: bold;
  margin-bottom: 12pt;
}

h1 { font-size: 18pt; }
h2 { font-size: 16pt; }
h3 { font-size: 14pt; }
h4 { font-size: 12pt; }
h5 { font-size: 10pt; }
h6 { font-size: 10pt; }

p, ul {
  margin: 1.286em 0;
  line-height: 120%;
}

dl {
  margin: 0;
  line-height: 120%;
}

pre {
  font-family: "Bitstream Vera Sans Mono", "Courier New", Courier, monospace;
  font-weight: normal !important;
  font-style: normal !important;
  font-size: 10pt !important;
  margin: 0px;
  overflow-x: auto;
}

strong {
  color: #000;
}

/* Pretty Tables */

.table table {
  border: 1px solid #036;
  padding: 0;
  border-collapse: collapse;
}

.table tr {
  border: 0;
  padding: 0;
}

.table tr:nth-child(even) {
  background-color: #EEE;
}

.table th {
  padding: 5px;
  background: #036;
  color: white;
  text-align: center;
  font-weight: bold;
}

.table td {
  border: 1px solid #CCC;
  padding: 5px;
}

h2.subtitle {
  font-variant: normal;
}

h3.author {
  font-variant: normal;
  margin-top: 2em;
  margin-bottom: 0;
}

.email {
}

.informalfigure {
  text-align: center;
}

.mediaobject {
  display: inline-block;
}

div.variablelist dd p {
  margin: 0;
  margin-top: 0;
  margin-bottom: 1em;
  margin-left: 1em;
}

span.term {
  font-weight:bold;
  padding-left:3px;
  padding-right:3px;
}

.caption p {
 margin: 0;
 text-align: right;
}

.navfooter table tr td {
  width: 33.33%;
}

.navfooter {
  margin-bottom: 1.0em;
}

.example .title {
  color: #248;
}

div .screenexample {
  background-color: #EEE;
  border:solid 1px #CCC;
  padding: 1.5em;
  color: black;
  border-radius: 15px;
}

pre.screen {
  white-space: pre;
  width: 100%;
  color: black;
  overflow-x: initial;
  white-space: pre-wrap;       /* css-3 */
  white-space: -moz-pre-wrap;  /* Mozilla, since 1999 */
  white-space: -pre-wrap;      /* Opera 4-6 */
  white-space: -o-pre-wrap;    /* Opera 7 */
  word-wrap: break-word;       /* Internet Explorer 5.5+ */
}

div.variablelist {
  padding-left: 20px;
}

.programlisting {
  color: black;
  font-weight: normal;
}

.programlisting .hl-tag {
  color: #881280;
  font-weight: normal;
}

.programlisting .hl-string {
  color: green;
  font-weight: bold;
}

.programlisting .hl-keyword {
  color: #009;
  font-weight: bold;
}

.programlisting .hl-attribute {
  color: #fc390e;
}

.programlisting .hl-value {
  color: #009;
}

.programlisting .hl-annotation {
  color: #880;
}

.programlisting .hl-comment {
  color: gray;
  font-style: italic;
}

.caution, .important, .note, .tip, .warning {
  padding: 10px;
  padding-left: 2.0em;
  margin: 10px;
  border-style: solid;
  border-width: 1px;
  border-top-width: 4px;
}

.caution { border-color: #c60; }
.caution h3 { color: #c60; }

.warning { border-color: #900; }
.warning h3 { color: #900; }

.important { border-color: #069; }
.important h3 { color: #069; }

.note { border-color: #909; }
.note h3 { color: #909; }

.tip { border-color: #090; }
.tip h3 { color: #090; }

.caution h3, .important h3, .note h3, .tip h3, .warning h3 {
  margin: 0px;
  margin-bottom: 0.5em;
}

.jetty-callout {
  background-color: #DEF;
  text-align: left; 
  border: thin solid #CCF; 
  padding: 4px; 
}

.jetty-callout h5 {
  font-size: 10pt;
  margin: 0;
  margin-bottom: 5px;
}

.jetty-callout .callout {
  font-weight: bold;
  font-style: normal;
}

.jetty-callout a {
  text-decoration: none;
}

.jetty-callout .website {
  color: #48f;
}

.jetty-callout p {
  font-style: oblique;
  margin: 0;
}

div.draft {
  background-color: #FF3300; 
  text-align: left; 
  font-size: 110%; 
  border:thin dotted blue; 
  padding: 4px;
  margin-top: 2.0em;
  margin-bottom: 2.0em;
  padding: 2.0em;
  color: white;
}

div.draft p {
  font-weight: bold;
  margin: 0;
}

div.draft h5 {
  margin: 0;
  margin-bottom: 1.0em;
  font-size: 150%;
}

div.draft a {
  color: yellow;
}

